// Copyright 2002 Adam Megacz, see the COPYING file for licensing [GPL]
package org.xwt;

import java.io.*;
import java.net.*;
import java.text.*;
import java.util.*;
import org.xwt.util.*;
import org.mozilla.javascript.*;

/** Singleton class that provides all functionality in the xwt.* namespace */
public final class XWT extends JSObject {

    public static final XWT singleton = new XWT();

    /** each key is a string representing a filename which the user has already given XWT permission to write to */
    private static Hashtable safeFiles = new Hashtable();

    public String getClassName() { return "XWT"; }
    private XWT() { setSeal(true); }

    public Object get(String name, Scriptable start) {
        if (name == null) return null;
        else if (name.equals("maxdim")) return new Integer(Short.MAX_VALUE);
        else if (name.equals("parseFloat")) return JSObject.defaultObjects.get("parseFloat", null);
        else if (name.equals("parseInt")) return JSObject.defaultObjects.get("parseInt", null);
        else if (name.equals("alt")) return Surface.alt ? Boolean.TRUE : Boolean.FALSE;
        else if (name.equals("control")) return Surface.control ? Boolean.TRUE : Boolean.FALSE;
        else if (name.equals("shift")) return Surface.shift ? Boolean.TRUE : Boolean.FALSE;
        else if (name.equals("date")) return date;
        else if (name.equals("listfonts")) return listfonts;
        else if (name.equals("regexp")) return regexp;
        else if (name.equals("sleep")) return sleep;
        else if (name.equals("yield")) return yield;
        else if (name.equals("newBrowserWindow")) return newBrowserWindow;
        else if (name.equals("textwidth")) return textwidth;
        else if (name.equals("textheight")) return textheight;
        else if (name.equals("newBox")) return newBox;
        else if (name.equals("soap")) return soap;
        else if (name.equals("xmlrpc")) return xmlrpc;
        else if (name.equals("origin")) return Main.origin;
        else if (name.equals("clipboard")) return Platform.getClipBoard();
        else if (name.equals("altKeyName")) return Platform.altKeyName();
        else if (name.equals("screenWidth")) return new Integer(Platform.getScreenWidth());
        else if (name.equals("screenHeight")) return new Integer(Platform.getScreenHeight());
        else if (name.equals("static")) return Static.getStatic("");
        else if (name.equals("theme")) return theme;
        else if (name.equals("openFile")) return openFile;
        else if (name.equals("saveFile")) return saveFile;
        else if (name.equals("saveFileAs")) return saveFileAs;
        else if (name.equals("utfEncode")) return utfEncode;
        else if (name.equals("fileSeparator")) return File.separator;
        else if (name.equals("homeDir")) return System.getProperty("user.home");
        else if (name.equals("tempDir")) return System.getProperty("java.io.tempdir");
        else if (name.equals("recursivePrintObject")) return recursivePrintObject;
        else if (name.equals("parseHTML")) return parseHTML;
        else if (name.equals("button")) {
            if (Surface.button1 && !Surface.button2 && !Surface.button3) return new Integer(1);
            else if (!Surface.button1 && Surface.button2 && !Surface.button3) return new Integer(1);
            else if (!Surface.button1 && !Surface.button2 && Surface.button3) return new Integer(1);
            else return new Integer(0);
        }
        else if (name.equals("println")) return println;
        else if (name.equals("math")) return org.xwt.util.JSObject.defaultObjects.get("Math", null);
        else return super.get(name, start);
    }

    public void put(String name, Scriptable start, Object value) {
        if (name == null) return;
        else if (name.equals("thread") && value != null && value instanceof Function) ThreadMessage.newthread((Function)value);
        else if (name.equals("clipboard")) Platform.setClipBoard(value.toString());
        else super.put(name, start, value);
    }


    // JSFunction Instances ///////////////////////////////////////////////////////////////////

    /** Helper class for defining functions. */
    private static abstract class JSFunction extends JSObject implements Function {
        JSFunction() { setSeal(true); }
        public Scriptable construct(Context cx, Scriptable scope, java.lang.Object[] args) { return null; }
    }

    private static final JSFunction newBrowserWindow = new JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args.length != 1 || args[0] == null) return null;
                Platform.newBrowserWindow(args[0].toString());
                return null;
            }
        };

    private static final JSFunction yield = new JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                sleep.call(cx, null, null, null);
                return null;
            }
        };

    private static final JSFunction println = new JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args.length == 1)
                    if (Log.on)
                        Log.log(cx.interpreterSourceFile, args[0] == null ? "null" : args[0].toString());
                return null;
            }
        };

    private static final JSFunction date = new JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                try { return Context.enter().newObject(org.xwt.util.JSObject.defaultObjects, "Date", args);
                } catch (Exception e) {
                    if (Log.on) Log.log(this, "Exception in Context.newObject() -- this should never happen");
                    if (Log.on) Log.log(this, e);
                    return null;
                }
            }
        };

    private static final JSFunction regexp = new JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                try { return Context.enter().newObject(org.xwt.util.JSObject.defaultObjects, "RegExp", args);
                } catch (Exception e) {
                    if (Log.on) Log.log(this, "Exception in Context.newObject() -- this should never happen");
                    if (Log.on) Log.log(this, e);
                    return null;
                }
            }
        };

    private static final JSFunction listfonts = new JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                return Context.enter().newArray(org.xwt.util.JSObject.defaultObjects,  Platform.listFonts());
            }
        };

    private static final JSFunction theme = new JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args.length != 2) return null;
                if (args[0] == null || args[1] == null) return null;
                Template.retheme(args[0].toString(), args[1].toString());
                return null;
            }
        };

    private static final JSFunction xmlrpc = new JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args.length != 1 || args[0] == null) return null;
                return new XMLRPC(args[0].toString(), "");
            }
        };

    private static final JSFunction soap = new JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args.length == 1 && args[0] != null) return new SOAP(args[0].toString(), "", null, null);
                else if (args.length == 2 && args[0] != null && args[1] != null)
                    return new SOAP(args[0].toString(), "", args[1].toString(), null);
                else if (args.length == 3 && args[0] != null && args[1] != null && args[2] != null)
                    return new SOAP(args[0].toString(), "", args[1].toString(), args[2].toString());
                else return null;
            }
        };

    private static final JSFunction textwidth = new JSFunction() {
            public Object call(Context cx, Scriptable scope, Scriptable thisObj, Object[] args) throws JavaScriptException {
                if (args.length < 1 || args.length > 2) return null;
                if (args[0] == null || (args.length == 2 && args[1] == null)) return null;
                String font = args.length == 1 ? Platform.getDefaultFont() : args[0].toString();
                String text = args.length == 1 ? args[0].toString() : args[1].toString();
                XWF xwf = XWF.getXWF(font);
                if (xwf == null) return new Integer(Platform.stringWidth(font, text));
                else return new Integer(xwf.stringWidth(text));
            }
        };


    private static final JSFunction textheight = new JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args.length > 1) return null;
                if (args.length == 1 && args[0] == null) return null;
                String font = args.length == 0 || args[0] == null ? Platform.getDefaultFont() : args[0].toString();
                XWF xwf = XWF.getXWF(font);
                if (xwf == null) return new Integer(Platform.getMaxAscent(font) + Platform.getMaxDescent(font));
                else return new Integer(xwf.getMaxAscent() + xwf.getMaxDescent());
            }
        };

    private static final JSFunction newBox = new JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                Box ret = new Box(args.length == 0 || args[0] == null ? "box" : args[0].toString(), Template.defaultImportList);
                for(int i=1; i<args.length; i++)
                    if (args[i] instanceof Box)
                        ret.put(ret.numChildren(), null, (Box)args[i]);
                for(int i=1; i<args.length; i++)
                    if (args[i] instanceof Scriptable && !(args[i] instanceof Box)) {
                        Scriptable s = (Scriptable)args[i];
                        Object[] keys = s.getIds();

                        // FIXME: need to ensure that this is putGlobally(), but still run traps...
                        for(int j=0; j<keys.length; j++) ret.put(keys[j].toString(), null, s.get(keys[j].toString(), s));
                    }
                return ret;
            }
        };

    private static final JSFunction sleep = new JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args != null && (args.length != 1 || args[0] == null)) return null;
                int i = args == null ? 0 : SpecialBoxProperty.stoi(args[0].toString());

                Thread thread = Thread.currentThread();
                if (!(thread instanceof ThreadMessage)) {
                    if (Log.on) Log.log(this, "cannot sleep() or yield() in the foreground thread");
                    return null;
                }
                ThreadMessage mythread = (ThreadMessage)thread;
                mythread.done.release();

                if (i > 0) try { Thread.sleep(i); } catch (Exception e) { }
                
                MessageQueue.add(mythread);
                mythread.go.block();
                return null;
            }
        };

    private static final JSFunction openFile = new JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args.length != 1) return null;
                String file = Platform.fileDialog(args[0].toString(), false);
                return file == null ? null : new ByteStream(file);
            }
        };

    private static final JSFunction saveFile = new JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args.length != 2) return null;
                if (!(args[1] instanceof ByteStream)) return null;
                String file = args[0].toString();
                if (safeFiles.get(Platform.isCaseSensitive() ? file : file.toLowerCase()) == null) {
                    file = Platform.fileDialog(file, true);
                    if (file == null) return null;
                    safeFiles.put(Platform.isCaseSensitive() ? file : file.toLowerCase(), new Object());
                }
                try {
                    ((ByteStream)args[1]).writeTo(new FileOutputStream(file));
                    return null;
                } catch (IOException e) {
                    if (Log.on) Log.log(ByteStream.class, "IO Exception while writing a ByteStream to a file");
                    if (Log.on) Log.log(ByteStream.class, e);
                    throw new JavaScriptException("error while writing a ByteStream to a file");
                }
            }
        };

    private static final JSFunction saveFileAs = new JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args.length != 2) return null;
                if (!(args[1] instanceof ByteStream)) return null;
                String file = args[0].toString();
                file = Platform.fileDialog(file, true);
                if (file == null) return null;
                safeFiles.put(Platform.isCaseSensitive() ? file : file.toLowerCase(), new Object());
                try {
                    ((ByteStream)args[1]).writeTo(new FileOutputStream(file));
                    return null;
                } catch (IOException e) {
                    if (Log.on) Log.log(ByteStream.class, "IO Exception while writing a ByteStream to a file");
                    if (Log.on) Log.log(ByteStream.class, e);
                    throw new JavaScriptException("error while writing a ByteStream to a file");
                }
            }
        };

    private static final JSFunction utfEncode = new JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args == null || args.length != 1) return null;
                return new ByteStream(args[0].toString().getBytes());
            }
        };

    
    private static final JSFunction parseHTML = new JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args == null || args.length != 1 || args[0] == null) return null;
                try {
                    if (args[0] instanceof ByteStream) {
                        return HTML.parseReader(new InputStreamReader(((ByteStream)args[0]).getInputStream()));
                    } else {
                        return HTML.parseReader(new StringReader(args[0].toString()));
                    }
                } catch (IOException e) {
                    if (Log.on) Log.log(HTML.class, "IO Exception while parsing HTML");
                    if (Log.on) Log.log(HTML.class, e);
                    throw new JavaScriptException("error while parsing HTML");
                }
            }
        };
    
    private static void recurse(String indent, String name, Object o, Context cx) {
        if (!name.equals("")) name += " : ";

        if (o == null) {
            Log.log(cx.interpreterSourceFile, indent + name + "<null>");

        } else if (o instanceof NativeArray) {
            Log.log(cx.interpreterSourceFile, indent + name + "<array>");
            NativeArray na = (NativeArray)o;
            for(int i=0; i<na.jsGet_length(); i++)
                recurse(indent + "  ", i + "", na.get(i, null), cx);

        } else if (!(o instanceof NativeDate) && (o instanceof JSObject || o instanceof ScriptableObject)) {
            Log.log(cx.interpreterSourceFile, indent + name + "<object>");
            Scriptable s = (Scriptable)o;
            Object[] keys = s.getIds();
            for(int i=0; i<keys.length; i++)
                recurse(indent + "  ", keys[i].toString(),
                        keys[i] instanceof Integer ? s.get(((Integer)keys[i]).intValue(), null) : s.get(keys[i].toString(), null), cx);

        } else {
            Log.log(cx.interpreterSourceFile, indent + name + o);

        }
    }

    private static final JSFunction recursivePrintObject = new JSFunction() {
            public Object call(Context cx, Scriptable thisObj, Scriptable ctorObj, Object[] args) throws JavaScriptException {
                if (args == null || args.length != 1) return null;
                recurse("", "", args[0], cx);
                return null;
            }
        };

}





