// Copyright 2002 Adam Megacz, see the COPYING file for licensing [LGPL]
package org.xwt.plat;

import org.xwt.*;
import org.xwt.util.*;
import java.util.*;
import java.io.*;

/** Platform specific code for GCJ-compiled Win32 binaries */
public class Win32 extends GCJ {

    // Initialization ////////////////////////////////////////////////////////////////////////////

    // Win32 often asks for a DC/Handle when it doesn't really need one
    static int desktop_handle = 0;
    static int desktop_dc = 0;

    // Cursors
    static int wait_cursor = 0;
    static int default_cursor = 0;
    static int crosshair_cursor = 0;
    static int text_cursor = 0;
    static int move_cursor = 0;
    static int sizenesw_cursor = 0;
    static int sizens_cursor = 0;
    static int sizenwse_cursor = 0;
    static int sizewe_cursor = 0;
    static int hand_cursor = 0;
    
    /** reverse lookup from hwnd to Win32Surface */
    public static Hashtable hwndToWin32SurfaceMap = new Hashtable();

    /** lets us know that natInit() is finished */
    public static Semaphore messagePumpStarted = new Semaphore();

    /** ThreadId of the message pump thread, used to send it messages */
    public static int messagePumpThread = 0;

    public static native String getTempPath();
    public static native void natInit();

    protected native String _fileDialog(String suggestedFileName, boolean write);

    public Win32() { }

    public void init() {
        String logfile = getTempPath() + "xwt-log.txt";
        try {
            PrintStream ps = new PrintStream(new FileOutputStream(logfile));
            System.setOut(ps);
            System.setErr(ps);
        } catch (Throwable e) {
            criticalAbort("Exception while attempting to redirect logging to " + logfile + " -- " + e);
        }

        new Thread() { public void run() { natInit(); } }.start();
        messagePumpStarted.block();
        fontList = new String[fontListVec.size()];
        fontListVec.toArray(fontList);
        fontListVec = null;
    }


    // Font Handling ////////////////////////////////////////////////////////////////////////////

    // FEATURE: query the registry for the user's default font
    protected String _getDefaultFont() { return "dialog8"; }
    protected int _getMaxAscent(String font) { return getFont(font).maxAscent; }
    protected int _getMaxDescent(String font) { return getFont(font).maxDescent; }
    protected native int _stringWidth(String font, String text);

    // methods/members used to enumerate platform fonts on startup
    public static Vector fontListVec = new Vector();
    public static String[] fontList = null;
    protected String[] _listFonts() { return fontList; }
    public static void addFont(String name, int height, boolean italic, boolean bold) {
        fontListVec.addElement(name.replace(' ', '_').toLowerCase() + "" + height + (italic ? "i" : "") + (bold ? "b" : ""));
    }

    static Hash fontCache = new Hash();
    public static class Win32Font {
        int hfont;
        int maxAscent;
        int maxDescent;
    }

    /** takes a parsed font and finds the closest platform-specific font */
    static native Win32Font mapFont(Platform.ParsedFont pf);

    /** takes an unparsed font and finds the closest platform-specific font */
    static Win32Font getFont(String font) {
        Win32Font ret = (Win32Font)fontCache.get(font);
        if (ret != null) return ret;

        Platform.ParsedFont pf = new Platform.ParsedFont(font);
        if (pf.name.equals("serif")) pf.name = "Times New Roman";
        else if (pf.name.equals("sansserif")) pf.name = "Arial";
        else if (pf.name.equals("monospace")) pf.name = "Courier New";
        else if (pf.name.equals("dialog")) pf.name = "Arial";
        else if (pf.name.equals("tty")) pf.name = "FixedSys";

        ret = mapFont(pf);
        fontCache.put(font, ret);
        return ret;
    }


    // Implementation of Platform methods /////////////////////////////////////////////////////////

    protected boolean _needsAutoClick() { return true; }
    protected String getDescriptiveName() { return "GCJ Win32 Binary"; }
    protected Surface _createSurface(Box b, boolean framed) { return new Win32Surface(b, framed); }
    protected DoubleBuffer _createDoubleBuffer(int w, int h, Surface owner) { return new Win32DoubleBuffer(w, h, (Win32Surface)owner); }
    protected Picture _createPicture(int[] b, int w, int h) { return new Win32Picture(b, w, h); }
    protected native int _getScreenWidth();
    protected native int _getScreenHeight();
    protected boolean _supressDirtyOnResize() { return false; }
    protected native void _criticalAbort(String message);
    protected native String _getClipBoard();
    protected native void _setClipBoard(String s);
    protected boolean _isCaseSensitive() { return false; }

    private native void __detectProxy(String[] container);

    protected synchronized HTTP.ProxyInfo _detectProxy() {

        String[] container = new String[] { null, null, null };
        if (Log.on) Log.log(this, "accessing Win32 registry");
        __detectProxy(container);
        if (container[2] == null && container[0] == null) {
            if (Log.on) Log.log(this, "no proxy settings in the Win32 registry");
            return null;
        }
        
        if (Log.on) Log.log(this, "PAC Script URL: " + container[2]);
        if (Log.on) Log.log(this, "Proxy Server String: " + container[0]);
        if (Log.on) Log.log(this, "Proxy Override String: " + container[1]);

        HTTP.ProxyInfo ret = new HTTP.ProxyInfo();
        if (container[2] != null) {
            ret.proxyAutoConfigFunction = HTTP.ProxyInfo.getProxyAutoConfigFunction(container[2]);
            if (ret.proxyAutoConfigFunction != null) return ret;
        }

        if (container[0] == null) return null;
        StringTokenizer st = new StringTokenizer(container[0], ";", false);
        while(st.hasMoreTokens()) try {
            String s = st.nextToken().trim();
            String protocol, host;
            if (s.indexOf(':') == -1) {
                continue;
            } else if (s.indexOf("://") != -1) {
                protocol = s.substring(0, s.indexOf("://"));
                s = s.substring(s.indexOf("://") + 3);
                host = s.substring(0, s.indexOf(':'));
            } else if (s.indexOf('=') == -1) {
                protocol = "http";
                host = s.substring(0, s.indexOf(':'));
            } else {
                protocol = s.substring(0, s.indexOf('='));
                host = s.substring(s.indexOf('=') + 1, s.indexOf(':'));
            }
            int port = Integer.parseInt(s.substring(s.indexOf(':') + 1));
            if (protocol.equals("http")) {
                ret.httpProxyHost = host;
                ret.httpProxyPort = port;
            } else if (protocol.equals("https")) {
                ret.httpsProxyHost = host;
                ret.httpsProxyPort = port;
            } else if (protocol.equals("socks")) {
                ret.socksProxyHost = host;
                ret.socksProxyPort = port;
            }
        } catch (NumberFormatException nfe) { }

        if (container[1] != null) {
            st = new StringTokenizer(container[1], ";", false);
            ret.excluded = new String[st.countTokens()];
            for(int i=0; st.hasMoreTokens(); i++) ret.excluded[i] = st.nextToken();
        }
        return ret;
    }

    protected native boolean _newBrowserWindow_(String url);
    protected void _newBrowserWindow(String url) {
        if (!_newBrowserWindow_(url))
            if (Log.on) Log.log(this, "ShellExecuteEx() failed trying to open url " + url);
    }

    // Win32Surface ////////////////////////////////////////////////////////////////////////////

    public static class Win32Surface extends Surface {

        /** used to block while waiting for the message pump thread to create a hwnd for us */
        public Semaphore hwndCreated = new Semaphore();

        /** nothing more than a method version of WndProc, so we can access instance members/methods */
        public native int WndProc(int hwnd, int imsg, int wparam, int lparam);

        /** true iff the mouse is inside this window; used to determine if we should capture the mouse */
        boolean inside = false;

        /** true iff we have 'captured' the mouse with SetCapture() */
        boolean captured = false;

        public int hwnd = -1;
        public int hdc = -1;

        public int current_cursor = default_cursor;

        /** used to restore the cursor immediately before ReleaseCapture() */
        public int previous_cursor = 0;

        public native void natInit(boolean framed);
        public Win32Surface(Box b, final boolean framed) {
            super(b);
            natInit(framed);
            hwndToWin32SurfaceMap.put(new Integer(hwnd), this);
        }

        public void syncCursor() {
            if (cursor.equals("default")) current_cursor = default_cursor;
            else if (cursor.equals("wait")) current_cursor = wait_cursor;
            else if (cursor.equals("crosshair")) current_cursor = crosshair_cursor;
            else if (cursor.equals("text")) current_cursor = text_cursor;
            else if (cursor.equals("move")) current_cursor = move_cursor;
            else if (cursor.equals("hand")) current_cursor = hand_cursor;
            else if (cursor.equals("east") || cursor.equals("west")) current_cursor = sizewe_cursor;
            else if (cursor.equals("north") || cursor.equals("south")) current_cursor = sizens_cursor;
            else if (cursor.equals("northwest") || cursor.equals("southeast")) current_cursor = sizenwse_cursor;
            else if (cursor.equals("northeast") || cursor.equals("southwest")) current_cursor = sizenesw_cursor;
            postCursorChange();
        }

        public native void finalize();
        public native void postCursorChange();
        public native void toBack();
        public native void toFront();
        public native void _setMinimized(boolean m);
        public native void setInvisible(boolean i);
        public native void _setMaximized(boolean m);
        public native void setSize(int w, int h);
        public native void setLocation(int x, int y);
        public native void setTitleBarText(String s);
        public native void setIcon(Picture p);
        public native void _dispose();
        public native void blit(DoubleBuffer s, int sx, int sy, int dx, int dy, int dx2, int dy2);
    }


    // Win32Picture ////////////////////////////////////////////////////////////////////////////

    public static class Win32Picture implements Picture {
        int w = 0, h = 0;
        int[] data = null;

        /** the Win32 bitmap version of this Picture */
        int hbitmap = -1;

        /** dc of the bitmap */
        int hdc = -1;

        /** true iff this Picture has translucent regions */
        boolean hasalpha = false;

        /** true iff this Picture has transparent regions but no translucent regions */
        boolean hasmask = true;

        /** if hasmask, this mask indicates which regions are transparent */
        int hmask = -1;

        /** dc of the mask */
        int maskdc = -1;

        public int getWidth() { return w; };
        public int getHeight() { return h; };
        public int[] getData() { return data; }
        public native void natInit();
        public Win32Picture(int[] data, int w, int h) { this.w = w; this.h = h; this.data = data; natInit(); }
    }
    

    // Win32DoubleBuffer //////////////////////////////////////////////////////////////////////////

    public static class Win32DoubleBuffer implements DoubleBuffer {

        int w = 0;
        int h = 0;

        int clipx1 = 0;
        int clipy1 = 0;
        int clipx2 = 0;
        int clipy2 = 0;

        int hdc = -1;
        int hbitmap = -1;

        public int getHeight() { return h; }
        public int getWidth() { return w; }

        public native void natInit();
        public Win32DoubleBuffer(int w, int h, Win32Surface owner) {
            this.w = w;
            this.h = h;
            clipx2 = w;
            clipy2 = h;
            natInit();
        }

        public native void setClip(int x, int y, int x2, int y2);
        public native void fillRect(int x, int y, int x2, int y2, int color);
        public native void drawString(String font, String text, int x, int y, int color);
        public native void drawPicture(Picture source, int dx1, int dy1, int dx2, int dy2, int sx1, int sy1, int sx2, int sy2);
        public native void finalize();
        public void drawPicture(Picture source, int x, int y) {
            drawPicture(source, x, y, x + source.getWidth(), y + source.getHeight(), 0, 0, source.getWidth(), source.getHeight());
        }

    }

}
                                    
